//
//  ProcessJsonData.swift
//  The purpose of this class is to process json data.
//  Fog of Text
//
//  Created by Joe Mccormack on 7/19/20.
//  Copyright © 2020 Joe Mccormack. All rights reserved.
//

import Foundation

/* This struct is defined in HandleJsonData.swift and only commented here for quick reference.
struct Pages: Codable {
    let pg: Int
    let paragraphs: [Paragraphs]
}
struct Paragraphs: Codable {
    let p: Int
    let words: [String]
}
*/

class ProcessJsonDataClass {
    
    /* Define non-alphanumeric characters to filter */
    /* var nonAlpha: Array = ["~", "`", "!", "@", "#", "$", "%", "^", "&", "*", "(", ")", "-", "_", "+", "=", "{", "[", "]", "}", "\\", "|", ":", "'", "\"", "<", ",", ">", ".", "?", "/"] */
    var nonAlpha: Array = ["~", "^", "\"", "\\"]
    
    /* Filter text string so that it contains only alpha-numeric characters */
    func filterWord(currentWord: String) -> String {
        var currentWordFiltered: String = ""
        for character in currentWord {
            var found:Int = 0
            for a:String in nonAlpha[0..<nonAlpha.count] {
                if String(character) == a {
                    found = 1
                }
            }
            if found == 0 {
                currentWordFiltered += String(character)
            }
        }
        return currentWordFiltered
    }
    
    /* Set random page, paragraph and word for encoding */
    var pageRnd: Int = 0
    var paragraphRnd: Int = 0
    var wordRnd: Int = 0
    var wordRndTxt: String = ""
    func generateRandomCoordsAndWord() {
        pageRnd = 0; paragraphRnd = 0; wordRnd = 0; wordRndTxt = "";
        var tmpNumber: Int = 0
        tmpNumber = loadedJSONData.count - 1
        pageRnd = Int.random(in: 0...tmpNumber)
        tmpNumber = loadedJSONData[pageRnd].paragraphs.count - 1
        paragraphRnd = Int.random(in: 0...tmpNumber)
        tmpNumber = loadedJSONData[pageRnd].paragraphs[paragraphRnd].words.count - 1
        wordRnd = Int.random(in: 0...tmpNumber)
        wordRndTxt = filterWord(currentWord: String(loadedJSONData[pageRnd].paragraphs[paragraphRnd].words[wordRnd]).lowercased())
        //wordRndTxt = String(loadedJSONData[pageRnd].paragraphs[paragraphRnd].words[wordRnd]).lowercased()
    }
    
    /* Encode the text string */
    func encodeStr(rawText: String) -> String {
        var encodedStr: String = ""
        var rawTextClean:String = rawText.trimmingCharacters(in: .whitespacesAndNewlines)
        /* Increase complexity to compensate for double whitespace split nil values since Swift does not have RegEx functionality to do it in an easier way */
        var rawWords: [String] = []
        var holdSwiftsHand: Array = rawTextClean.components(separatedBy: " ")
        for sw in 0..<holdSwiftsHand.count {
            if holdSwiftsHand[sw].count > 0 {
                rawWords.append(String(holdSwiftsHand[sw]))
            }
        }
        /* Determine upper recursion limit based on word count of source */
        var countUpperLimit: Int = 0
        var straightCount: Int = 0
        var numberOfPages: Int = loadedJSONData.count
        for c in 0..<numberOfPages {
            var numberOfParagraphsOfPage: Int = loadedJSONData[c].paragraphs.count
            for cp in 0..<numberOfParagraphsOfPage {
                straightCount += loadedJSONData[c].paragraphs[cp].words.count
            }
        }
        countUpperLimit = (straightCount * 2) + (straightCount / 2)
        /* Step through each word in text block */
        var currentWordIndex: Int = -1
        var rawWordsSize: Int = rawWords.count
        for rawStep in 0..<rawWordsSize {
            /* Set word to match for encoding */
            currentWordIndex += 1
            var currentWord: String = String(rawWords[currentWordIndex])
            /* Preserve the leading or trailing non-alphanumeric character, if present */
            var currentWordFirstCharNonAlpha: Bool = false
            var currentWordFirstChar: String = String(currentWord[currentWord.index(currentWord.startIndex, offsetBy: 0)])
            var currentWordLastCharNonAlpha: Bool = false
            var compensateForSwiftGag: Int = currentWord.count - 1
            var currentWordLastChar: String = String(currentWord[currentWord.index(currentWord.startIndex, offsetBy: compensateForSwiftGag)])
            for n:String in nonAlpha[0..<nonAlpha.count] {
                if n == currentWordFirstChar {
                    currentWordFirstCharNonAlpha = true
                }
                if n == currentWordLastChar {
                    currentWordLastCharNonAlpha = true
                }
            }
            if currentWordFirstCharNonAlpha == false {
                currentWordFirstChar = " "
            }
            if currentWordLastCharNonAlpha == false {
                currentWordLastChar = " "
            }
            /* Filter word so that it contains only alpha-numeric characters */
            var currentWordFiltered: String = filterWord(currentWord: currentWord.lowercased())
            /* If nothing was returned, the string only contained non-alphanumeric characters.  Preserve it and skip additional processing. */
            if currentWordFiltered.count == 0 {
                currentWordFirstCharNonAlpha = true
                currentWordFirstChar = currentWord
                currentWordLastCharNonAlpha = false
                currentWordLastChar = ""
                encodedStr += "~^" + String(currentWordFirstChar) + "~^" + "0" + "~^" + "0" + "~^" + "0" + "~^" + String(currentWordLastChar) + "~^" + "0" + "~^" + "-1" + "^~"
            }
            else {
                /* Continue processing - loop maximum of countUpperLimit to match word */
                var wordMatched: Int = 0
                for _ in 0..<countUpperLimit {
                    /* Randomize page, paragraph, word to attempt to match string to */
                    generateRandomCoordsAndWord()
                    /* Determine if there is a match and, if so, record the source location */
                    if wordRndTxt == currentWordFiltered {
                        wordMatched = 1
                        /* Build encoded string for word match */
                        encodedStr += "~^" + String(currentWordFirstChar) + "~^" + String(pageRnd) + "~^" + String(paragraphRnd) + "~^" + String(wordRnd) + "~^" + String(currentWordLastChar) + "~^" + "1" + "~^" + "-1" + "^~"
                        /* Exit for loop */
                        break
                    }
                }
                if wordMatched == 0 {
                    /* Attept to iterate word character by character to see if each character is found in sufficient, random source words */
                    var characterBucket: String = ""
                    var lastCharacterFound: String = ""
                    var foundCharacter: Int = 0
                    var preserveOriginalWord: Int = 0
                    var swift_wc: Int = 0
                    for _ in 0..<currentWordFiltered.count {
                        foundCharacter = 0
                        var currentCharacter: String = String(currentWordFiltered[currentWordFiltered.index(currentWordFiltered.startIndex, offsetBy: swift_wc)])
                        for _ in 0..<countUpperLimit {
                            /* Randomize page, paragraph, word to attempt to match a character of */
                            generateRandomCoordsAndWord()
                            /* Iterate through source word to see if it contains current character */
                            var swift_swc: Int = 0
                            for _ in 0..<wordRndTxt.count {
                                var wordRndTxtCharAt: String = String(wordRndTxt[wordRndTxt.index(wordRndTxt.startIndex, offsetBy: swift_swc)])
                                if wordRndTxtCharAt == currentCharacter {
                                    foundCharacter = 1
                                    lastCharacterFound = wordRndTxtCharAt
                                    var preCharacter: String = " "
                                    var postCharacter: String = " "
                                    if (swift_wc == 0) {
                                        preCharacter = currentWordFirstChar
                                    }
                                    if swift_wc == (currentWordFiltered.count - 1) {
                                        if currentWordLastChar.trimmingCharacters(in: .whitespacesAndNewlines).count == 0 {
                                            /* End marker for column when decoding when non-alphanumeric is missing */
                                            currentWordLastChar = "`"
                                        }
                                        postCharacter = currentWordLastChar
                                    }
                                    characterBucket += "~^" + String(preCharacter) + "~^" + String(pageRnd) + "~^" + String(paragraphRnd) + "~^" + String(wordRnd) + "~^" + String(postCharacter) + "~^" + "0" + "~^" + String(swift_swc) + "^~"
                                    /* Exit for loop */
                                    break
                                }
                                swift_swc += 1
                            }
                            /* If current character was found, exit to find next character in source words */
                            if foundCharacter == 1 {
                                /* Exit for loop */
                                break
                            }
                        }
                        /* Current character could not be matched so exit */
                        if lastCharacterFound != currentCharacter {
                            preserveOriginalWord = 1
                            /* Exit for loop */
                            break
                        }
                    swift_wc += 1
                    }
                    /* If character not found in any source words, preserve the original word */
                    if preserveOriginalWord == 1 {
                        /* Preserve original word */
                        encodedStr += "~^" + String(currentWord) + "~^" + "0" + "~^" + "0" + "~^" + "0" + "~^" + " " + "~^" + "0" + "~^" + "-1" + "^~"
                    }
                    else {
                        /* All characters were found in source words, use encoding */
                        encodedStr += characterBucket
                    }
                }
            }
        }
        
        return encodedStr
    }
    
    /* Decode the text string */
    var decodedStrFailMsg: String = ""
    func decodeStr(rawText: String) -> String {
        decodedStrFailMsg = ""
        var decodedStr: String = ""
        var wordSequenceStarted: Int = 0
        var rawTextClean:String = rawText.trimmingCharacters(in: .whitespacesAndNewlines)
        var rawWords:Array = rawTextClean.components(separatedBy: "^~")
        for rw in 0..<rawWords.count {
            var originalWordOrPrefix: String = ""
            var pageNumber: Int = 0
            var paragraphNumber: Int = 0
            var wordNumber: Int = 0
            var originalPostfix: String = ""
            var isWord: Int = 0
            var isCharacterOfWord: Int = -1
            var rawWordSet: Array = rawWords[rw].components(separatedBy: "~^")
            if rawWordSet.count == 8 {
                originalWordOrPrefix = String(rawWordSet[1])
                pageNumber = Int(rawWordSet[2])!
                paragraphNumber = Int(rawWordSet[3])!
                wordNumber = Int(rawWordSet[4])!
                originalPostfix = String(rawWordSet[5])
                isWord = Int(rawWordSet[6])!
                isCharacterOfWord = Int(rawWordSet[7])!
                if isCharacterOfWord < 0 {
                    if wordSequenceStarted == 1 {
                        wordSequenceStarted = 0
                    }
                    if isWord > 0 {
                        /* Encoded word */
                        if originalWordOrPrefix.trimmingCharacters(in: .whitespacesAndNewlines).count > 0 {
                            decodedStr += String(originalWordOrPrefix)
                        }
                        decodedStr += String(loadedJSONData[pageNumber].paragraphs[paragraphNumber].words[wordNumber].lowercased())
                        if originalPostfix.trimmingCharacters(in: .whitespacesAndNewlines).count > 0 {
                            decodedStr += originalPostfix
                        }
                        decodedStr += " "
                    }
                    else {
                        /* Non-encoded word */
                        decodedStr += originalWordOrPrefix + " "
                    }
                }
                else if isCharacterOfWord >= 0  {
                    /* Encoded word as characters */
                    if wordSequenceStarted == 0 {
                        if originalWordOrPrefix.trimmingCharacters(in: .whitespacesAndNewlines).count > 0 {
                            decodedStr += originalWordOrPrefix
                        }
                    }
                    var sourceWord: String = String(loadedJSONData[pageNumber].paragraphs[paragraphNumber].words[wordNumber].lowercased())
                    decodedStr += String(sourceWord[sourceWord.index(sourceWord.startIndex, offsetBy: isCharacterOfWord)])
                    /* Last character of word */
                    if originalPostfix == "`" {
                        decodedStr += " "
                    }
                    else if originalPostfix.trimmingCharacters(in: .whitespacesAndNewlines).count > 0 {
                        decodedStr += originalPostfix
                        decodedStr += " "
                    }
                    wordSequenceStarted = 1
                }
                else {
                    /* Encoded string format problem; most likely the source was modified */
                    decodedStrFailMsg = "Encoding to decode does not appear to be structured correctly, contains unsupported data, or is missing."
                }
            }
        }
        
        return decodedStr
    }

}
